@@ -532,6 +532,22 @@ message_remove_filter_cb (GeditMessageBus *bus,
 	gedit_file_browser_widget_remove_filter (data->widget, id);
 }
 
+static void
+message_extend_context_menu_cb (GeditMessageBus *bus,
+				GeditMessage    *message,
+				GeditWindow     *window)
+{
+	WindowData *data;
+	GeditMenuExtension *ext;
+
+	data = get_window_data (window);
+
+	ext = gedit_file_browser_widget_extend_context_menu (data->widget);
+
+	g_object_set (message, "extension", ext, NULL);
+	g_object_unref (ext);
+}
+
 static void
 message_up_cb (GeditMessageBus *bus,
 	       GeditMessage    *message,
@@ -675,6 +691,11 @@ register_methods (GeditWindow            *window,
 	                            MESSAGE_OBJECT_PATH,
 	                            "remove_filter");
 
+	gedit_message_bus_register (bus,
+	                            GEDIT_TYPE_FILE_BROWSER_MESSAGE_EXTEND_CONTEXT_MENU,
+	                            MESSAGE_OBJECT_PATH,
+	                            "extend_context_menu");
+
 	gedit_message_bus_register (bus,
 	                            GEDIT_TYPE_MESSAGE,
 	                            MESSAGE_OBJECT_PATH,
@@ -726,6 +747,7 @@ register_methods (GeditWindow            *window,
 	BUS_CONNECT (bus, set_markup, data);
 	BUS_CONNECT (bus, add_filter, window);
 	BUS_CONNECT (bus, remove_filter, data);
+	BUS_CONNECT (bus, extend_context_menu, window);
 
 	BUS_CONNECT (bus, up, data);
 	BUS_CONNECT (bus, history_back, data);
@@ -1039,4 +1061,4 @@ gedit_file_browser_messages_unregister (GeditWindow *window)
 
 	window_data_free (window);
 }
-/* ex:ts=8:noet: */
+/* ex:set ts=8 noet: */
